package xyz.lychee.lagfixer.libs.adventure.audience;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Objects;
import java.util.UUID;
import java.util.function.Consumer;
import java.util.function.Predicate;
import java.util.stream.Collector;
import xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar;
import xyz.lychee.lagfixer.libs.adventure.chat.ChatType;
import xyz.lychee.lagfixer.libs.adventure.chat.SignedMessage;
import xyz.lychee.lagfixer.libs.adventure.identity.Identified;
import xyz.lychee.lagfixer.libs.adventure.identity.Identity;
import xyz.lychee.lagfixer.libs.adventure.inventory.Book;
import xyz.lychee.lagfixer.libs.adventure.pointer.Pointered;
import xyz.lychee.lagfixer.libs.adventure.resource.ResourcePackInfo;
import xyz.lychee.lagfixer.libs.adventure.resource.ResourcePackInfoLike;
import xyz.lychee.lagfixer.libs.adventure.resource.ResourcePackRequest;
import xyz.lychee.lagfixer.libs.adventure.resource.ResourcePackRequestLike;
import xyz.lychee.lagfixer.libs.adventure.sound.Sound;
import xyz.lychee.lagfixer.libs.adventure.sound.SoundStop;
import xyz.lychee.lagfixer.libs.adventure.text.Component;
import xyz.lychee.lagfixer.libs.adventure.text.ComponentLike;
import xyz.lychee.lagfixer.libs.adventure.title.Title;
import xyz.lychee.lagfixer.libs.adventure.title.TitlePart;
import xyz.lychee.lagfixer.libs.annotations.NotNull;
import xyz.lychee.lagfixer.libs.annotations.ApiStatus.ScheduledForRemoval;

public interface Audience extends Pointered {
   static @NotNull Audience empty() {
      return EmptyAudience.INSTANCE;
   }

   static @NotNull Audience audience(final @NotNull Audience... audiences) {
      int length = audiences.length;
      if (length == 0) {
         return empty();
      } else {
         return (Audience)(length == 1 ? audiences[0] : audience((Iterable)Arrays.asList(audiences)));
      }
   }

   static @NotNull ForwardingAudience audience(final @NotNull Iterable<? extends Audience> audiences) {
      return () -> {
         return audiences;
      };
   }

   static @NotNull Collector<? super Audience, ?, ForwardingAudience> toAudience() {
      return Audiences.COLLECTOR;
   }

   default @NotNull Audience filterAudience(final @NotNull Predicate<? super Audience> filter) {
      return filter.test(this) ? this : empty();
   }

   default void forEachAudience(final @NotNull Consumer<? super Audience> action) {
      action.accept(this);
   }

   @ForwardingAudienceOverrideNotRequired
   default void sendMessage(final @NotNull ComponentLike message) {
      this.sendMessage(message.asComponent());
   }

   default void sendMessage(final @NotNull Component message) {
      this.sendMessage(message, MessageType.SYSTEM);
   }

   /** @deprecated */
   @Deprecated
   @ForwardingAudienceOverrideNotRequired
   @ScheduledForRemoval(
      inVersion = "5.0.0"
   )
   default void sendMessage(final @NotNull ComponentLike message, final @NotNull MessageType type) {
      this.sendMessage(message.asComponent(), type);
   }

   /** @deprecated */
   @Deprecated
   @ForwardingAudienceOverrideNotRequired
   @ScheduledForRemoval(
      inVersion = "5.0.0"
   )
   default void sendMessage(final @NotNull Component message, final @NotNull MessageType type) {
      this.sendMessage(Identity.nil(), message, type);
   }

   /** @deprecated */
   @Deprecated
   @ForwardingAudienceOverrideNotRequired
   default void sendMessage(final @NotNull Identified source, final @NotNull ComponentLike message) {
      this.sendMessage(source, message.asComponent());
   }

   /** @deprecated */
   @Deprecated
   @ForwardingAudienceOverrideNotRequired
   default void sendMessage(final @NotNull Identity source, final @NotNull ComponentLike message) {
      this.sendMessage(source, message.asComponent());
   }

   /** @deprecated */
   @Deprecated
   @ForwardingAudienceOverrideNotRequired
   default void sendMessage(final @NotNull Identified source, final @NotNull Component message) {
      this.sendMessage(source, message, MessageType.CHAT);
   }

   /** @deprecated */
   @Deprecated
   @ForwardingAudienceOverrideNotRequired
   default void sendMessage(final @NotNull Identity source, final @NotNull Component message) {
      this.sendMessage(source, message, MessageType.CHAT);
   }

   /** @deprecated */
   @Deprecated
   @ForwardingAudienceOverrideNotRequired
   @ScheduledForRemoval(
      inVersion = "5.0.0"
   )
   default void sendMessage(final @NotNull Identified source, final @NotNull ComponentLike message, final @NotNull MessageType type) {
      this.sendMessage(source, message.asComponent(), type);
   }

   /** @deprecated */
   @Deprecated
   @ForwardingAudienceOverrideNotRequired
   @ScheduledForRemoval(
      inVersion = "5.0.0"
   )
   default void sendMessage(final @NotNull Identity source, final @NotNull ComponentLike message, final @NotNull MessageType type) {
      this.sendMessage(source, message.asComponent(), type);
   }

   /** @deprecated */
   @Deprecated
   @ScheduledForRemoval(
      inVersion = "5.0.0"
   )
   default void sendMessage(final @NotNull Identified source, final @NotNull Component message, final @NotNull MessageType type) {
      this.sendMessage(source.identity(), message, type);
   }

   /** @deprecated */
   @Deprecated
   @ScheduledForRemoval(
      inVersion = "5.0.0"
   )
   default void sendMessage(final @NotNull Identity source, final @NotNull Component message, final @NotNull MessageType type) {
   }

   default void sendMessage(final @NotNull Component message, final ChatType.@NotNull Bound boundChatType) {
      this.sendMessage(message, MessageType.CHAT);
   }

   @ForwardingAudienceOverrideNotRequired
   default void sendMessage(final @NotNull ComponentLike message, final ChatType.@NotNull Bound boundChatType) {
      this.sendMessage(message.asComponent(), boundChatType);
   }

   default void sendMessage(final @NotNull SignedMessage signedMessage, final ChatType.@NotNull Bound boundChatType) {
      Component content = signedMessage.unsignedContent() != null ? signedMessage.unsignedContent() : Component.text(signedMessage.message());
      if (signedMessage.isSystem()) {
         this.sendMessage((Component)content);
      } else {
         this.sendMessage((Identity)signedMessage.identity(), (Component)content, MessageType.CHAT);
      }

   }

   @ForwardingAudienceOverrideNotRequired
   default void deleteMessage(final @NotNull SignedMessage signedMessage) {
      if (signedMessage.canDelete()) {
         this.deleteMessage((SignedMessage.Signature)Objects.requireNonNull(signedMessage.signature()));
      }

   }

   default void deleteMessage(final SignedMessage.@NotNull Signature signature) {
   }

   @ForwardingAudienceOverrideNotRequired
   default void sendActionBar(final @NotNull ComponentLike message) {
      this.sendActionBar(message.asComponent());
   }

   default void sendActionBar(final @NotNull Component message) {
   }

   @ForwardingAudienceOverrideNotRequired
   default void sendPlayerListHeader(final @NotNull ComponentLike header) {
      this.sendPlayerListHeader(header.asComponent());
   }

   default void sendPlayerListHeader(final @NotNull Component header) {
      this.sendPlayerListHeaderAndFooter((Component)header, (Component)Component.empty());
   }

   @ForwardingAudienceOverrideNotRequired
   default void sendPlayerListFooter(final @NotNull ComponentLike footer) {
      this.sendPlayerListFooter(footer.asComponent());
   }

   default void sendPlayerListFooter(final @NotNull Component footer) {
      this.sendPlayerListHeaderAndFooter((Component)Component.empty(), (Component)footer);
   }

   @ForwardingAudienceOverrideNotRequired
   default void sendPlayerListHeaderAndFooter(final @NotNull ComponentLike header, final @NotNull ComponentLike footer) {
      this.sendPlayerListHeaderAndFooter(header.asComponent(), footer.asComponent());
   }

   default void sendPlayerListHeaderAndFooter(final @NotNull Component header, final @NotNull Component footer) {
   }

   @ForwardingAudienceOverrideNotRequired
   default void showTitle(final @NotNull Title title) {
      Title.Times times = title.times();
      if (times != null) {
         this.sendTitlePart(TitlePart.TIMES, times);
      }

      this.sendTitlePart(TitlePart.SUBTITLE, title.subtitle());
      this.sendTitlePart(TitlePart.TITLE, title.title());
   }

   default <T> void sendTitlePart(final @NotNull TitlePart<T> part, final @NotNull T value) {
   }

   default void clearTitle() {
   }

   default void resetTitle() {
   }

   default void showBossBar(final @NotNull BossBar bar) {
   }

   default void hideBossBar(final @NotNull BossBar bar) {
   }

   default void playSound(final @NotNull Sound sound) {
   }

   default void playSound(final @NotNull Sound sound, final double x, final double y, final double z) {
   }

   default void playSound(final @NotNull Sound sound, final Sound.@NotNull Emitter emitter) {
   }

   @ForwardingAudienceOverrideNotRequired
   default void stopSound(final @NotNull Sound sound) {
      this.stopSound(((Sound)Objects.requireNonNull(sound, "sound")).asStop());
   }

   default void stopSound(final @NotNull SoundStop stop) {
   }

   @ForwardingAudienceOverrideNotRequired
   default void openBook(final Book.@NotNull Builder book) {
      this.openBook(book.build());
   }

   default void openBook(final @NotNull Book book) {
   }

   @ForwardingAudienceOverrideNotRequired
   default void sendResourcePacks(final @NotNull ResourcePackInfoLike first, final ResourcePackInfoLike... others) {
      this.sendResourcePacks(ResourcePackRequest.addingRequest(first, others));
   }

   @ForwardingAudienceOverrideNotRequired
   default void sendResourcePacks(final @NotNull ResourcePackRequestLike request) {
      this.sendResourcePacks(request.asResourcePackRequest());
   }

   default void sendResourcePacks(final @NotNull ResourcePackRequest request) {
   }

   @ForwardingAudienceOverrideNotRequired
   default void removeResourcePacks(final @NotNull ResourcePackRequestLike request) {
      this.removeResourcePacks(request.asResourcePackRequest());
   }

   @ForwardingAudienceOverrideNotRequired
   default void removeResourcePacks(final @NotNull ResourcePackRequest request) {
      List<ResourcePackInfo> infos = request.packs();
      if (infos.size() == 1) {
         this.removeResourcePacks(((ResourcePackInfo)infos.get(0)).id());
      } else if (infos.isEmpty()) {
         return;
      }

      UUID[] otherReqs = new UUID[infos.size() - 1];

      for(int i = 0; i < otherReqs.length; ++i) {
         otherReqs[i] = ((ResourcePackInfo)infos.get(i + 1)).id();
      }

      this.removeResourcePacks(((ResourcePackInfo)infos.get(0)).id(), otherReqs);
   }

   @ForwardingAudienceOverrideNotRequired
   default void removeResourcePacks(final @NotNull ResourcePackInfoLike request, final @NotNull ResourcePackInfoLike... others) {
      UUID[] otherReqs = new UUID[others.length];

      for(int i = 0; i < others.length; ++i) {
         otherReqs[i] = others[i].asResourcePackInfo().id();
      }

      this.removeResourcePacks(request.asResourcePackInfo().id(), otherReqs);
   }

   default void removeResourcePacks(final @NotNull Iterable<UUID> ids) {
      Iterator<UUID> it = ids.iterator();
      if (it.hasNext()) {
         UUID id = (UUID)it.next();
         UUID[] others;
         if (!it.hasNext()) {
            others = new UUID[0];
         } else if (ids instanceof Collection) {
            others = new UUID[((Collection)ids).size() - 1];

            for(int i = 0; i < others.length; ++i) {
               others[i] = (UUID)it.next();
            }
         } else {
            List<UUID> othersList = new ArrayList();

            while(it.hasNext()) {
               othersList.add((UUID)it.next());
            }

            others = (UUID[])othersList.toArray(new UUID[0]);
         }

         this.removeResourcePacks(id, others);
      }
   }

   default void removeResourcePacks(final @NotNull UUID id, final @NotNull UUID... others) {
   }

   default void clearResourcePacks() {
   }
}
